"""Various functions for TTRPG Gamesmasters to help with dice rolls."""

from .dice import Dice as d  # noqa: N813
from .manydice import LazyRollTable, PoolComparison, lazyroll
from .statblock import StatBlock, statblock
